package com.ringlord.odfind;

import java.awt.HeadlessException;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.io.IOException;

import java.util.List;
import java.util.ArrayList;

import javax.swing.SwingUtilities;

import com.ringlord.GPL3;

/**
 * <p>The main entry point for this program.</p>
 *
 * <p>Invokes the {@linkplain GUI Graphic User Interface class}
 *    if a GUI is available. On headless sytems, those without
 *    a graphics card or a graphics system loaded) an error
 *    message is printed to the console.</p>
 *
 * @author K Udo Schuermann
**/
class Main
{
  public static final String VERSION = "2.0";
  public static final String VERDATE = "26-May-2012";

  public static void main( final String[] args )
  {
    boolean showUsage = false;
    boolean needsNewline = false;

    if( args.length == 0 )
      {
        try
          {
            SwingUtilities.invokeLater( new GUI() );
            return;
          }
        catch( final InternalError noGUI )
          {
          }
        catch( final HeadlessException noGUI )
          {
            System.err.println( "Non-graphical environment?" );
          }
        showHeader();
        System.err.println( "ERROR: Non-graphical environment?" );
        showUsage = true;
        needsNewline = true;
      }

    boolean showVersion = false;
    boolean showLicense = false;
    String lastFlag = "";
    final List<String> queries = new ArrayList<String>();
    final List<String> roots = new ArrayList<String>();
    File passwordFile = null;
    boolean interactiveConsole = false;
    boolean checkEncryptedFiles = true;
    boolean ignoreCase = false;
    boolean recurse = false;
    boolean acceptFlags = true;
    for( int i=0; i<args.length; i++ )
      {
        final String arg = args[i];
        if( acceptFlags &&
            arg.startsWith("-") )
          {
            lastFlag = arg;
            if( arg.equals("--") )
              {
                // everything starting with '-' from now on is a
                // normal argument, not treated as a flag
                acceptFlags = false;
              }
            else if( arg.equals("-R") ||
                     arg.equals("--recurse") )
              {
                recurse = true;
              }
            else if( arg.equals("-i") ||
                     arg.equals("--case-insensitive") )
              {
                ignoreCase = true;
              }
            else if( arg.equals("-s") )
              {
                // already recorded 'lastFlag'
              }
            else if( arg.equals("-d") )
              {
                // already recorded 'lastFlag'
              }
            else if( arg.equals("-V") ||
                     arg.equals("--version") )
              {
                showVersion = true;
              }
            else if( arg.equals("-I") ||
                     arg.equals("--interactive") )
              {
                interactiveConsole = true;
                checkEncryptedFiles = true;
              }
            else if( arg.equals("-P") ||
                     arg.equals("--password-file") )
              {
                if( passwordFile != null )
                  {
                    showHeader();
                    System.err.println( "ERROR: Only one -P/--password-file is supported" );
                    showUsage = true;
                    needsNewline = true;
                  }
                else if( ++i < args.length )
                  {
                    passwordFile = new File( args[i] );
                    checkEncryptedFiles = true;
                  }
                else
                  {
                    showHeader();
                    System.err.println( "ERROR: Expected filename following "+arg );
                    showUsage = true;
                    needsNewline = true;
                  }
              }
            else if( arg.equals("-H") ||
                     arg.equals("--help") )
              {
                showUsage = true;
              }
            else if( arg.equals("-L") ||
                     arg.equals("--license") )
              {
                showLicense = true;
              }
            else
              {
                showHeader();
                System.err.println( "ERROR: Unrecognized flag: "+arg );
                showUsage = true;
                needsNewline = true;
              }
          }
        else if( lastFlag.equals("-s") ||
                 lastFlag.equals("--search") ||
                 lastFlag.equals("--query") )
          {
            queries.add( arg );
          }
        else if( lastFlag.equals("-d") ||
                 lastFlag.equals("--directory") ||
                 lastFlag.equals("--dir") )
          {
            roots.add( arg );
          }
        else
          {
            // Emulating the basic argument order of grep:
            // first non-prefixed argument is a keyword
            // second non-prefixed argument is the location (directory)
            if( queries.isEmpty() )
              {
                queries.add( arg );
              }
            else if( roots.isEmpty() )
              {
                roots.add( arg );
              }
            else
              {
                showHeader();
                System.err.println( "ERROR: Unrecognized argument: "+arg );
                System.err.println( "       Did you forget a -s or -d flag when specifying" );
                System.err.println( "       multiple search queries and/or directories?" );
                showUsage = true;
                needsNewline = true;
              }
          }
      }

    if( showLicense )
      {
        showHeader();
        System.out.println( GPL3.LICENSE );
        System.out.println();
        return;
      }

    if( showVersion )
      {
        System.out.println( "ODFind "+VERSION+" ("+VERDATE+")" );
        return;
      }

    if( ! showUsage )
      {
        if( queries.isEmpty() )
          {
            showHeader();
            System.err.println( "ERROR: Need at least one search/query string" );
            showUsage = true;
            needsNewline = true;
          }
        if( roots.isEmpty() )
          {
            showHeader();
            System.err.println( "ERROR: Need at least one directory to search" );
            showUsage = true;
            needsNewline = true;
          }
      }

    if( showUsage )
      {
        if( needsNewline )
          {
            System.err.println();
          }
        showHeader();
        System.err.println( usage );
        return;
      }

    boolean passwordFileFailed = false;
    if( checkEncryptedFiles &&
        (passwordFile != null) )
      {
        try
          {
            final PasswordManager pwdMan = PasswordManager.getInstance();
            final BufferedReader f = new BufferedReader( new FileReader(passwordFile) );
            String s;
            while( (s = f.readLine()) != null )
              {
                pwdMan.add( s );
              }
            f.close();
          }
        catch( IOException x )
          {
            System.err.println( "ERROR: Cannot read passwords from file '"+passwordFile+"'" );
            passwordFileFailed = true;
          }
      }

    PasswordManager.getInstance().setCLI( checkEncryptedFiles,
                                          interactiveConsole );

    for( String root : roots )
      {
        System.err.println( "Searching "+root );

        final String[] q = new String[ queries.size() ];
        queries.toArray( q );

        final TreeWalker w = new TreeWalker( null );
        w.setQuery( q,
                    TreeWalker.QueryType.AUTO_DETECT,
                    ! ignoreCase );
        w.setRoot( new File(root).toPath(),
                   recurse );
        System.err.println( "Running query for "+queries+" on "+root );

        // The following will not ACTUALLY run in the background as we
        // are herewith invoking the method that would otherwise run
        // on a background thread. That is PRECISELY the intent. When
        // this method returns all files have been processed.
        w.doInBackground();

        PasswordManager.getInstance().finishCLI();
      }

    if( passwordFileFailed &&
        checkEncryptedFiles &&
        ! interactiveConsole )
      {
        System.err.println( "NOTE: Skipped encrypted files: Loading of "+
                            "passwords failed, and interactive mode was "+
                            "not enabled." );
      }
  }

  private static boolean headerShown;
  private static void showHeader()
  {
    if( ! headerShown )
      {
        headerShown = true;
        System.err.println( "ODFind "+VERSION+" ("+VERDATE+")" );
        System.err.println( "Copyright (C) 2008-2012 Ringlord Technologies" );
        System.err.println( "Copyright (C) 2008-2012 K. Udo Schuermann" );
        System.err.println( "All rights reserved" );
        System.err.println();
        System.err.println( "This program comes with ABSOLUTELY NO WARRANTY. This is free" );
        System.err.println( "software, and you are welcome to redistribute it under certain" );
        System.err.println( "conditions. Run it with -L on the command line for details." );
        System.err.println();
      }
  }

  private static final String usage =
    "Usage:\n"+
    "  -s [ query [...] ]            What to search for. This can be a simple string\n"+
    "  --search [ query [...] ]      or a regular expression. The first value given\n"+
    "  --query [ query [...] ]       without a -s will be assumed to be a query.\n"+
    "\n"+
    "  -d [ path [...] ]             Where to search, a directory or filename. You\n"+
    "  --dir [ path [...] ]          probably want to specify -R (--recurse) also if\n"+
    "  --directory [ path [...] ]    want to search subdirectories, too. The second\n"+
    "                                value given without a -d will be assumed to be\n"+
    "                                a directory to search.\n"+
    "\n"+
    "  -i                            All of your queries will ignore all upper/lower\n"+
    "  --case-insensitive            case variations in both query and documents.\n"+
    "\n"+
    "  -R                            Recurse through subdirectories, too. Most of\n"+
    "  --recurse                     the time you probably want this.\n"+
    "\n"+
    "  -I                            Check encrypted files and enable interactive\n"+
    "                                collection of passwords on the Console.\n"+
    "\n"+
    "  -P [ filename ]               Preload a set of passwords from the given file\n"+
    "  --password-file [ filename ]  to enable checking of encrypted files. You may\n"+
    "                                also want to use -I to collect more passwords\n"+
    "                                when the given list does not decrypt a file.\n"+
    "\n"+
    "  -v                            Output (to stderr) information about the search\n"+
    "  --verbose                     process, as well as statistics at the end.\n"+
    "\n"+
    "  -L                            Print the GNU General Public License (GPL) v3\n"+
    "  --license                     under which ODFind is licensed. Note that the\n"+
    "                                source code for ODFind is stored inside the JAR\n"+
    "                                and this includes an Apache Ant build.xml file,\n"+
    "                                too, so you can rebuild the software easily!\n"+
    "\n"+
    "  -H                            This helpful message. More extensive help may\n"+
    "  --help                        be found at http://ringlord.com/odfind/\n"+
    "\n"+
    "  -V                            Print ODFind's version to stdout and exit.\n"+
    "  --version\n"+
    "\n"+
    "Without any command line options the interactive GUI will be invoked.\n"+
    "The command line parser is primitive: you cannot run options together.\n"+
    "Example: ... -R -i lenna $HOME/Documents\n"+
    "Searches recursively (-R) and without regard to upper/lower case (-i) for the\n"+
    "word 'lenna' in the directory tree 'Documents' within your home directory.\n"+
    "As only a single search key and single path is given, we omitted -s and -d but\n"+
    "could also have written it as ... -R -i -s lenna -d $HOME/Documents\n"+
    "If you want to search encrypted documents, too, then add -I and/or -P options.\n";
}
