package com.ringlord.odfind;

import java.io.File;
import java.io.IOException;

import java.nio.charset.Charset;

import java.nio.file.Path;
import java.nio.file.Files;
import java.nio.file.FileVisitor;
import java.nio.file.FileVisitResult;
import java.nio.file.FileVisitOption;
import java.nio.file.attribute.BasicFileAttributes;

import java.util.Set;
import java.util.HashSet;
import java.util.regex.Pattern;

import javax.swing.SwingWorker;

import com.ringlord.odf.Container;
import com.ringlord.odf.Entry;

/**
 * <p>Walks the directory tree (recursively if configured to do so)
 * and searches each file encountered along the way. Matches are
 * reported to the given {@link ResultsPanel}, but encrypted files are
 * passed to the PasswordManager, who also gets a reference to the
 * ResultsPanel to be notified if the decrypted file is a match.</p>
 *
 * @author K. Udo Schuermann
 **/
public class TreeWalker
  extends SwingWorker<Void,Void>
{
  /**
   * Construct a TreeWalker with a reference to the ResultsPanel,
   * which is where results are displayed. If the ResultsPanel is null
   * then user feedback will be limited, and restricted to the console
   * only.
   **/
  public TreeWalker( final ResultsPanel results )
  {
    super();
    this.results = results;
  }

  public void setRoot( final Path root,
                       final boolean isRecursive )
  {
    if( isRunning )
      {
        throw new IllegalStateException( "Cannot modify state on running search" );
      }
    this.root = root;
    this.isRecursive = isRecursive;
  }
  public void setQuery( final String query,
                        final QueryType type,
                        final boolean isCaseSensitive )
  {
    setQuery( new String[]{query},
              type,
              isCaseSensitive );
  }
  public void setQuery( final String[] queries,
                        final QueryType type,
                        final boolean isCaseSensitive )
  {
    if( isRunning )
      {
        throw new IllegalStateException( "Cannot modify state on running search" );
      }
    switch( type )
      {
      case AUTO_DETECT:
        try
          {
            int options = Pattern.MULTILINE | Pattern.UNICODE_CASE | Pattern.DOTALL;
            if( ! isCaseSensitive )
              {
                options |= Pattern.CASE_INSENSITIVE;
              }
            this.regex = new Pattern[ queries.length ];
            for( int i=0; i<queries.length; i++ )
              {
                this.regex[i] = Pattern.compile( queries[i],
                                                 options );
              }
          }
        catch( Exception x )
          {
            this.regex = null;
            this.substrings = queries;
          }
        break;

      case SUBSTRING:
        this.substrings = queries;
        break;

      case REGEX:
        int options = Pattern.MULTILINE | Pattern.UNICODE_CASE | Pattern.DOTALL;
        if( ! isCaseSensitive )
          {
            options |= Pattern.CASE_INSENSITIVE;
          }
        this.regex = new Pattern[ queries.length ];
        for( int i=0; i<queries.length; i++ )
          {
            this.regex[i] = Pattern.compile( queries[i],
                                             options );
          }
        break;

      default:
        throw new IllegalArgumentException( "Unhandled QueryType: "+type );
      }

    this.isCaseSensitive = isCaseSensitive;
  }

  public Void doInBackground()
  {
    ResultsPanel.ResultComponent r;
    if( results != null ) // do we have a results GUI for output?
      {
        // Construct the ResultComponent where search progress is
        // displayed for the user, and matching files are listed.
        final String[] queryText;
        if( substrings != null )
          {
            queryText = substrings;
          }
        else
          {
            queryText = new String[ regex.length ];
            for( int i=0; i<regex.length; i++ )
              {
                queryText[i] = regex[i].toString();
              }
          }
        r = results.addSearchPanel( this,
                                    queryText );
      }
    else
      {
        // There is no output (other than the console)
        r = null;
      }

    try
      {
        final Set<FileVisitOption> walkerOptions = new HashSet<FileVisitOption>();
        if( r == null )
          {
            walkerOptions.add( FileVisitOption.FOLLOW_LINKS );
          }
        else
          {
            if( Settings.get("search-follow-symlinks",false) )
              {
                // @@@ for the CLI version this must be controlled by a command line option!
                walkerOptions.add( FileVisitOption.FOLLOW_LINKS );
              }
          }
        this.inspector = new Inspector( r,
                                        isRecursive,
                                        Settings.get("search-raw-xml",false),
                                        Settings.get("search-all-parts",false) );
        Files.walkFileTree( root,
                            walkerOptions,
                            Integer.MAX_VALUE,
                            inspector );
      }
    catch( IOException x )
      {
        x.printStackTrace();
      }
    finally
      {
        if( r != null )
          {
            r.done();
          }
      }

    return null;
  }

  public void stop()
  {
    if( inspector != null )
      {
        inspector.stop();
      }
  }

  /**
   * What kind of query does the user indicate the search string to be
   **/
  public enum QueryType
  {
    AUTO_DETECT,
    SUBSTRING,
    REGEX;
  }



  class Inspector
    implements FileVisitor<Path>
  {
    Inspector( final ResultsPanel.ResultComponent r,
               final boolean isRecursive,
               final boolean searchRawXML,
               final boolean searchAllParts )
    {
      super();
      this.r = r;
      this.recurse = isRecursive;
      this.searchRawXML = searchRawXML;
      this.searchAllParts = searchAllParts;
    }

    void stop()
    {
      this.result = FileVisitResult.TERMINATE;
    }
    public FileVisitResult visitFile( final Path file,
                                      BasicFileAttributes attribs )
      throws IOException
    {
      searchFile( file );
      return result;
    }
    public FileVisitResult visitFileFailed( final Path file,
                                            final IOException x )
    {
      return result;
    }
    public FileVisitResult preVisitDirectory( final Path directory,
                                              final BasicFileAttributes attribs )
    {
      if( recurse ||
          (recursionLevel == 0) )
        {
          recursionLevel++;
          if( r == null )
            {
              System.err.print( directory.toString() + "[K[1G" );
            }
          else
            {
              r.scanning( directory );
            }
          return result;
        }
      return FileVisitResult.SKIP_SUBTREE;
    }
    public FileVisitResult postVisitDirectory( final Path directory,
                                               final IOException x )
    {
      --recursionLevel;
      return result;
    }

    /**
     * <p>Test whether the given {@link byte[] data} (coming from the
     * named {@link Path file}) is a match for our search criteria.
     * If it is, the file will be recorded as a match; if not, it is
     * simply ignored and nothing further is said about it.</p.
     *
     * @param data The pre-extracted, possibly decrypted, and inflated
     * data from the file (in other words, the original data that we
     * can search directly). Usually this is XML.
     *
     * @param file The file to be searched; this parameter is the
     * primary part of the report when the file does match the search
     * criteria.
     *
     * @param password The password that decrypted the original data,
     * or null if the file was plain text (no password was required).
     *
     * @return true if the file matched the search criteria (and this
     * was reported) or false if it did not match.
     **/
    boolean search( final byte[] data,
                    final Path file,
                    final String password )
    {
      // @@@ search the name, too (this should be an option)

      final String s;
      if( searchRawXML )
        {
          s = new String(data,UTF8);
        }
      else
        {
          s = new String(data,UTF8).replace( "<\\.+>", "" );
        }
      boolean isMatch = false;
      if( substrings != null )
        {
          for( int i=0; i<substrings.length; i++ )
            {
              if( isCaseSensitive )
                {
                  isMatch = s.toUpperCase().indexOf(substrings[i]) > -1;
                }
              else
                {
                  isMatch = s.indexOf(substrings[i]) > -1;
                }
              if( isMatch )
                {
                  break;
                }
            }
        }
      else
        {
          for( int i=0; i<regex.length; i++ )
            {
              if( isMatch = regex[i].matcher( s ).find() )
                {
                  break;
                }
            }
        }

      if( isMatch )
        {
          if( r != null )
            {
              r.matches( file, password );
            }
          else
            {
              System.out.println( (password != null
                                   ? "[C]"
                                   : "[P]")+
                                  " "+file );
            }
          return true;
        }
      return false;
    }

    /**
     * <p>Initiate processing of the given {@link Path file},
     * determining along the way whether the file can be processed
     * directly (it is not encrypted) or needs to be processed through
     * the {@link PasswordManager} because collected passwords need to
     * be checked on the document, and bad passwords re-checked (over
     * and over).</p.
     *
     * @param file The file to be searched; a search will quietly fail
     * if the file is not an ODF container.
     **/
    private void searchFile( final Path file )
    {
      // optionally exclude files that do not match the filename pattern

      // optionally exclude files that have the wrong filename extension

      try
        {
          final Container odf = new Container( file.toFile() );
          // If it's not an ODF container then it's at least a zip file!

          // @@@ if 'searchAllParse' is true then scan ALL items in the
          // @@@ container, not just 'content.xml'

          final Entry content = odf.get( "content.xml" );
          if( content == null )
            {
              return; // no content.xml => doesn't look like an ODF file
            }

          if( content.isEncrypted() )
            {
              PasswordManager.getInstance().add( this, r, file, odf, content );
              // PasswordManager will quickly check the file against
              // all collected passwords before closing the Container
              // and asking the user for another password. In other
              // words *WE* do not close the file, PasswordManager
              // will take it from here.
            }
          else
            {
              final byte[] data = content.data();
              // The search method will either record the fact that
              // the file matched, or it will silently ignore it and
              // say nothing about it.
              search( data, file, null );
              odf.close();
            }
        }
      // A large number of exceptions can occur, all of which mean
      // basically that the data is not accessible. Hence, we catch a
      // generic 'Exception' here:
      catch( Exception x )
        {
          // The file is not an ODF container, or something else went
          // horribly wrong. Whatever the specific case, there is no
          // way for us to actually search the content of this file,
          // and complaining too loudly is going to serve nobody.
        }
    }

    private int recursionLevel;
    private FileVisitResult result = FileVisitResult.CONTINUE;
    //
    private final boolean recurse;
    private final boolean searchRawXML;
    private final boolean searchAllParts;
    private final ResultsPanel.ResultComponent r;
  }

  private boolean isRunning;
  private Inspector inspector;
  //
  private Path root;
  private boolean isRecursive;
  private boolean isCaseSensitive;
  // Only one of the following should be non-null
  private String[] substrings;
  private Pattern[] regex;
  // ------------------------------
  private final ResultsPanel results;
  private static final Charset UTF8 = Charset.forName("UTF-8");
}
